/*
        earray.h -- elastic array
        (C) Copyright 1995  John Webster Small
        All rights reserved

        TYPE must be a selectible class having default
        and copy initializer constructors, overloaded
        assignment operator, and overloaded stream
        insertion and extraction operators.
*/

#ifndef earray_h
#define earray_h

#ifndef accessor_h
  #include "accessor.h"
#endif

template <class TYPE>
class ElasticArray
{
    Accessor<TYPE > A;
    static TYPE * newArray(size_t arrayCells)
        {
          assert(arrayCells);
          return new TYPE[arrayCells];
        }
  public:
    ElasticArray() : A() {}
    ElasticArray(const ElasticArray<TYPE > & ea)
        : A(ea.A) {}
    ElasticArray(size_t arrayCells)
        : A(newArray(arrayCells),arrayCells) {}
    ElasticArray(TYPE * array, size_t arrayCells)
        : A(array,arrayCells) {}
    ElasticArray<TYPE >& operator()(size_t arrayCells)
        {
          (void) A(newArray(arrayCells),arrayCells);
          return *this;
        }
    ElasticArray<TYPE >& operator()
        (TYPE * array, size_t arrayCells)
        {
          (void) A(array,arrayCells);
          return *this;
        }
    ElasticArray<TYPE >& realloc(size_t size_);
    ElasticArray<TYPE >&
        operator=(const ElasticArray<TYPE > & ea)
            { A = ea.A; return *this; }
    size_t size() { return A.arrayCells(); }
    TYPE & operator[](size_t n) { return A[n]; }
    TYPE * cpy(size_t dst, const TYPE * T, size_t n);
    TYPE * move(size_t dst, size_t src, size_t n);
    TYPE * set(size_t dst, TYPE& t, size_t n);
    friend ostream& operator<<
        (ostream& os, ElasticArray<TYPE > & ea)
        { return os << ea.A << endl; }
    friend istream& operator>>
        (istream& is, ElasticArray<TYPE > & ea)
        { is >> ea.A; is.get(); return is; }
};

template <class TYPE>
ElasticArray<TYPE >&
ElasticArray<TYPE >:: realloc(size_t size_)
{
    TYPE * T = 0;
    if (size_)  {
        T = new TYPE[size_];
        if (T)  {
            size_t n = ((A.arrayCells() < size_)?
                A.arrayCells() : size_);
            if (n)  {
                TYPE * _T = (TYPE *)A;
                assert(_T);
                while (n--)
                    T[n] = _T[n];
            }
        }
    }
    A(T,size_);
    return *this;
}

template <class TYPE>
TYPE * ElasticArray<TYPE >::
    cpy(size_t dst, const TYPE * T, size_t n)
{
    assert(dst + n >= dst && dst + n >= n);
    assert(dst + n <= A.arrayCells());
    TYPE * _T = (TYPE *)A;
    assert (_T && T);
    _T += dst;
    while (n--)
        *_T++ = *T++;
    return ((TYPE *)A) + dst;
}

template <class TYPE>
TYPE * ElasticArray<TYPE >::
    move(size_t dst, size_t src, size_t n)
{
    assert(dst + n >= dst && dst + n >= n);
    assert(dst + n <= A.arrayCells());
    assert(src + n >= src && src + n >= n);
    assert(src + n <= A.arrayCells());
    TYPE * D = ((TYPE *)A) + dst;
    TYPE * S = ((TYPE *)A) + src;
    assert (D);
    while (n--)
        *D++ = *S++;
    return ((TYPE *)A) + dst;
}

template <class TYPE>
TYPE * ElasticArray<TYPE >::
    set(size_t dst, TYPE& t, size_t n)
{
    assert(dst + n >= dst && dst + n >= n);
    assert(dst + n <= A.arrayCells());
    TYPE * _T = ((TYPE *)A);
    assert (_T);
    _T += dst;
    while (n--)
        *_T++ = t;
    return ((TYPE *)A) + dst;
}

#endif
