/*
        psingle.h -- allow only a single instance
          per <SCOPE,TYPE,ID> tuple!
        (C) Copyright 1995  John Webster Small
        All rights reserved

        TYPE must have a default constructor and
        overloaded stream insertion and extraction
        operators.
*/

#ifndef psingle_h
#define psingle_h

/* LINTLIBRARY */

#ifndef assert_h
  #include <assert.h>
  #ifndef assert_h
    #define assert_h
  #endif
#endif

#ifndef limits_h
  #include <limits.h>
  #ifndef limits_h
    #define limits_h
  #endif
#endif

#ifndef fstream_h
  #include <fstream.h>
  #ifndef fstream_h
    #define fstream_h
  #endif
#endif

#ifndef cstr_h
  #include "cstr.h"
#endif

#ifndef GLOBAL_SCOPE
  #define SCOPE_TYPE(scope) scope ## _SCOPE
  #define SCOPE_TYPEDEF(scope) \
    struct SCOPE_TYPE(scope) {}
  SCOPE_TYPEDEF(global);
  #define GLOBAL_SCOPE SCOPE_TYPE(global)
#endif

template <class SCOPE, class TYPE, int ID = 0>
struct PSingleton_traits
{
    static TYPE * newInstance() { return new TYPE; }
};

enum PSingletonOp
    { PSGLTN_RELEASE, PSGLTN_CREATE, PSGLTN_PRESENT };

template <class SCOPE, class TYPE, int ID = 0>
class PSingleton_Base
{
  public:
    const char * pathname
        (const char * filename = 0,
         const char * path = 0);
    unsigned long refs(int inc = 0) const;
    TYPE * instance
        (PSingletonOp op = PSGLTN_CREATE) const;
    PSingleton_Base(const char * filename,
        const char * path = 0)
        { (void) refs(1);
          (void) pathname(filename,path); }
    ~PSingleton_Base()
        { if (!refs(-1)) (void)
              instance(PSGLTN_RELEASE); }
};

template <class SCOPE, class TYPE, int ID>
const char *
PSingleton_Base<SCOPE,TYPE,ID>::
    pathname(const char * filename, const char * path)
{
    static cstr _pathname;
    if (filename)  {
        cstr tmp(path);
        tmp.cat(filename);
        if (_pathname.len())
            assert(!_pathname.cmp(tmp));
        else
            _pathname.cpy(tmp);
    }
    return _pathname;
}

template <class SCOPE, class TYPE, int ID>
unsigned long
PSingleton_Base<SCOPE,TYPE,ID>::refs(int inc) const
{
    static unsigned long _refs = 0;
    if (inc > 0)  {
        assert(_refs < ULONG_MAX);
        _refs++;
    }
    else if (inc < 0)  {
        assert(_refs);
        _refs--;
    }
    return _refs;
}

template <class SCOPE, class TYPE, int ID>
TYPE *
PSingleton_Base<SCOPE,TYPE,ID>::
    instance(PSingletonOp op) const
{
    static TYPE * _instance = 0;
    if (op != PSGLTN_RELEASE && !refs()) return 0;
    if (op == PSGLTN_RELEASE && _instance)  {
        ofstream os(pathname(),ios::out|ios::binary);
        if (os)  {
            os << *_instance;
            os.close();
        }
        delete _instance;
        _instance = 0;
    }
    if (op == PSGLTN_CREATE && !_instance)  {
        _instance = PSingleton_traits::newInstance();
        if (_instance)  {
            ifstream is(pathname(),ios::in|ios::binary);
            if (is)  {
                is >> *_instance;
                is.close();
            }
        }
    }
    return _instance;
};

template <class SCOPE, class TYPE, int ID = 0>
class PSingleton : PSingleton_Base<SCOPE,TYPE,ID>
{
  public:
    PSingleton(const char * filename,
        const char * path = 0)
        : PSingleton_Base<SCOPE,TYPE,ID>(filename,path)
        { assert(pathname()); }
    PSingleton_Base<SCOPE,TYPE,ID>::refs;
    unsigned long refs() const { return
        PSingleton_Base<SCOPE,TYPE,ID>::refs(); }
    operator TYPE * ()
        { return instance(PSGLTN_CREATE); }
    TYPE * ok() const
        { return instance(PSGLTN_PRESENT); }
    operator TYPE & () 
        { TYPE * T = instance(PSGLTN_CREATE);
          assert(T); return *T; }
    void release() { (void) instance(PSGLTN_RELEASE); }
};

#endif
